import * as consola from 'consola';
import z from 'zod';
import * as express from 'express';
import express__default, { Request } from 'express';
import { Kysely } from 'kysely';
import * as node_mocks_http from 'node-mocks-http';
import { Transporter } from 'nodemailer';
import JSONTransport from 'nodemailer/lib/json-transport';
import Mail from 'nodemailer/lib/mailer';
import SQLite from 'better-sqlite3';
import { Queue } from 'plainjobs';

type LogLevel = "error" | "warn" | "normal" | "info" | "debug" | "trace";
declare function getLogger(name?: string): consola.ConsolaInstance;

type InputConfig = {
    nodeEnv: "development" | "production" | "test";
    logger: {
        level?: LogLevel;
    };
    http?: {
        port?: number;
        staticPath?: string;
    };
    paths?: {
        routes?: string;
        commands?: string;
        jobs?: string;
        schedules?: string;
        schema?: string;
        assets?: string;
        out?: string;
        styles?: string;
        seed?: string;
        migrations?: string;
        database?: string;
        http?: string;
        queue?: string;
        mailer?: string;
        logger?: string;
    };
};
type Config = {
    nodeEnv: "development" | "production" | "test";
    logger: {
        level: LogLevel;
    };
    http: {
        port: number;
        staticPath: string;
    };
    paths: {
        routes: string;
        commands: string;
        jobs: string;
        schedules: string;
        assets: string;
        out: string;
        schema: string;
        styles: string;
        seed: string;
        migrations: string;
        database: string;
        http: string;
        queue: string;
        mailer: string;
        logger: string;
    };
};
declare function defineConfig(config: InputConfig): InputConfig;
declare function getConfig(): Config;

/** Return a random cuid2 id with an optional prefix.
 * @example
 * randomId("usr") // usr_1a2b3c4d
 * randomId() // 1a2b3c4d
 */
declare function randomId(prefix?: string): string;

declare function dev(): boolean;
declare function prod(): boolean;
declare function test(): boolean;
type Zod = typeof z;
declare function defineEnv<T>(f: (z: Zod) => z.Schema<T>): T;

interface PlainResponse {
    _tag: "PlainResponse";
    headers?: Record<string, string>;
    status?: number;
    body?: string | Promise<string>;
    htmlStream?: (rid: number | string) => JSX.Element;
}
/** Return a HTML response with Content-Type set to text/html. */
declare function html(html: Promise<string> | string, opts?: {
    status?: number;
}): PlainResponse;
/**
 * Stream a HTML response with Content-Type set to text/html.
 * Cache-Control is set to no-transform to prevent caching.
 */
declare function stream(htmlStream: (rid: number | string) => JSX.Element): PlainResponse;
/**
 * Redirect to a new path.
 * The status code defaults to 302.
 */
declare function redirect(path: string, opts?: {
    status?: number;
}): PlainResponse;
/** Return a JSON response with Content-Type set to application/json. */
declare function json(json: JSONSerializable, opts?: {
    status?: number;
}): PlainResponse;
/** Return a 404 response. */
declare function notFound(): PlainResponse;

interface HandlerArgs {
    req: express__default.Request;
    res: express__default.Response;
}
type JSONSerializable = string | number | boolean | null | JSONSerializable[] | {
    [key: string]: JSONSerializable;
};
type Handler = (args: HandlerArgs) => Promise<ExpressResponse | PlainResponse | JSX.Element | JSONSerializable>;
type ExpressResponse = () => void;
/**
 * Defines a handler function with proper typing.
 *
 * @param handler The handler function to be defined
 * @returns The typed handler function
 */
declare function defineHandler(handler: (args: HandlerArgs) => Promise<ExpressResponse | PlainResponse | JSX.Element | JSONSerializable>): Handler;

type FileRoute = {
    filePath: string;
    GET?: Handler;
    POST?: Handler;
};
/** Return an express router that serves files as routes form the given directory. */
declare function fileRouter$1(opts?: {
    cwd?: string;
    config?: Config;
}): Promise<{
    paths: string[];
    router: express__default.RequestHandler;
    routes: FileRoute[];
}>;

type GenericDatabase = Kysely<any>;
/**
 * Run a a function in a database transaction.
 * The transaction is automatically rolled back, even if the function doesn't throw an error.
 * Use during testing, to keep test cases isolated from each other.
 * */
declare function rollback<T>(db: Kysely<T>, fn: (db: Kysely<T>) => Promise<void>): Promise<void>;
declare function defineDatabase<T>(db: Kysely<T>): Kysely<T>;

declare function forceWWW(): express__default.RequestHandler;
declare function logging(): express__default.RequestHandler;
declare function rateLimit(opts?: {
    rateLimit?: {
        windowMs?: number;
        limit?: number;
        message?: string;
    };
}): express__default.RequestHandler;
declare function database({ database, }: {
    database: GenericDatabase;
}): express__default.RequestHandler;
declare function pendingMigrations(): express__default.RequestHandler;
declare function fileRouter(): Promise<express__default.RequestHandler>;
declare function loadConfig(): express__default.RequestHandler;
declare function id(): express__default.RequestHandler;
/** A collection of built-in express middleware. */
declare const middleware: {
    forceWWW: typeof forceWWW;
    logging: typeof logging;
    rateLimit: typeof rateLimit;
    database: typeof database;
    fileRouter: typeof fileRouter;
    id: typeof id;
    pendingMigrations: typeof pendingMigrations;
    loadConfig: typeof loadConfig;
};
/** Print all express routes to the console. Useful for debugging. */
declare function printRoutes(app: Express.Application): Promise<void>;
declare function defineHttp(handler: (config: Config) => Promise<express__default.Application>): (config: Config) => Promise<express__default.Application>;

/** Run a handler with a mock request, returning the response. Useful for testing. */
declare function testHandler(handler: Handler, req: Request, { database, }: {
    database?: GenericDatabase;
}): Promise<node_mocks_http.MockResponse<express.Response<any, Record<string, any>>>>;

/**
 * Return the HTTP static path for a given asset path.
 * Refer to the raw assets, not the transpiled assets.
 *
 * Example:
 * asset("scripts.ts") // returns "/public/scripts.js"
 */
declare function asset(path: string): string;

/**
 * This outbox traps sent emails during testing (NODE_ENV=test).
 * Access the outbox during testing to assert on sent emails.
 */
declare const outbox: JSONTransport.SentMessageInfo[];
/**
 * Define a mailer using nodemailer, that can be used to send emails.
 */
declare function defineMailer(config: {
    host: string;
    port: number;
    secure: boolean;
    auth: {
        user: string;
        pass: string;
    };
}): Transporter;
/** Send an email by providing a mailers instance. */
declare function sendMail(mailer: Transporter, mailOptions: Mail.Options): Promise<void>;

declare function defineSeed(f: (db: GenericDatabase) => Promise<void>): {
    run: (db: GenericDatabase) => Promise<void>;
};

type Job<T> = {
    name: string;
    run: ({ data }: {
        data: T;
    }) => Promise<void>;
};
declare function defineJob<T>(opts: {
    name: string;
    run: ({ data }: {
        data: T;
    }) => Promise<void>;
}): Job<T>;
type Schedule = {
    name: string;
    cron: string;
    run: () => Promise<void>;
};
declare function defineSchedule(opts: {
    name: string;
    cron: string;
    run: () => Promise<void>;
}): Schedule;
declare function defineQueue(opts: {
    connection: SQLite.Database;
}): Queue;
declare function work(queue: Queue, jobs: Record<string, Job<unknown>>, schedules: Record<string, Schedule>): Promise<void>;
declare function perform<T>(job: Job<T>, data?: T): Promise<void>;

export { type Config, type Handler, type Job, type Schedule, asset, defineConfig, defineDatabase, defineEnv, defineHandler, defineHttp, defineJob, defineMailer, defineQueue, defineSchedule, defineSeed, dev, fileRouter$1 as fileRouter, getConfig, getLogger, html, json, middleware, notFound, outbox, perform, printRoutes, prod, randomId, redirect, rollback, sendMail, stream, test, testHandler, work };
